@@ -49,39 +49,60 @@ import datetime
 #
 # Helper functions
 #
-def GetEccByType(handle, counterType, bitType):
-    try:
-        count = str(nvmlDeviceGetTotalEccErrors(handle, bitType, counterType))
-    except NVMLError as err:
-        count = handleError(err)
+def GetEccByType(handle, counterType, errorType):
+    strResult = ''
     
     try:
-        detail = nvmlDeviceGetDetailedEccErrors(handle, bitType, counterType)
-        deviceMemory = str(detail.deviceMemory)
-        registerFile = str(detail.registerFile)
-        l1Cache = str(detail.l1Cache)
-        l2Cache = str(detail.l2Cache)
+        deviceMemory = nvmlDeviceGetMemoryErrorCounter(handle, errorType, counterType,
+                                                       NVML_MEMORY_LOCATION_DEVICE_MEMORY)
     except NVMLError as err:
-        msg = handleError(err)
-        deviceMemory = msg
-        registerFile = msg
-        l1Cache = msg
-        l2Cache = msg
-    strResult = ''
-    strResult += '          <device_memory>' + deviceMemory + '</device_memory>\n'
-    strResult += '          <register_file>' + registerFile + '</register_file>\n'
-    strResult += '          <l1_cache>' + l1Cache + '</l1_cache>\n'
-    strResult += '          <l2_cache>' + l2Cache + '</l2_cache>\n'
+        deviceMemory = handleError(err)
+    strResult += '          <device_memory>' + str(deviceMemory) + '</device_memory>\n'
+    
+    try:
+        registerFile = nvmlDeviceGetMemoryErrorCounter(handle, errorType, counterType,
+                                                       NVML_MEMORY_LOCATION_REGISTER_FILE)
+    except NVMLError as err:
+        registerFile = handleError(err)
+    
+    strResult += '          <register_file>' + str(registerFile) + '</register_file>\n'
+    
+    try:
+        l1Cache = nvmlDeviceGetMemoryErrorCounter(handle, errorType, counterType,
+                                                  NVML_MEMORY_LOCATION_L1_CACHE)
+    except NVMLError as err:
+        l1Cache = handleError(err)
+    strResult += '          <l1_cache>' + str(l1Cache) + '</l1_cache>\n'
+    
+    try:
+        l2Cache = nvmlDeviceGetMemoryErrorCounter(handle, errorType, counterType,
+                                                  NVML_MEMORY_LOCATION_L2_CACHE)
+    except NVMLError as err:
+        l2Cache = handleError(err)
+    strResult += '          <l2_cache>' + str(l2Cache) + '</l2_cache>\n'
+    
+    try:
+        textureMemory = nvmlDeviceGetMemoryErrorCounter(handle, errorType, counterType,
+                                                        NVML_MEMORY_LOCATION_TEXTURE_MEMORY)
+    except NVMLError as err:
+        textureMemory = handleError(err)
+    strResult += '          <texture_memory>' + str(textureMemory) + '</texture_memory>\n'
+    
+    try:
+        count = str(nvmlDeviceGetTotalEccErrors(handle, errorType, counterType))
+    except NVMLError as err:
+        count = handleError(err)
     strResult += '          <total>' + count + '</total>\n'
+    
     return strResult
 
 def GetEccByCounter(handle, counterType):
     strResult = ''
     strResult += '        <single_bit>\n'
-    strResult += str(GetEccByType(handle, counterType, NVML_SINGLE_BIT_ECC))
+    strResult += str(GetEccByType(handle, counterType, NVML_MEMORY_ERROR_TYPE_CORRECTED))
     strResult += '        </single_bit>\n'
     strResult += '        <double_bit>\n'
-    strResult += str(GetEccByType(handle, counterType, NVML_DOUBLE_BIT_ECC))
+    strResult += str(GetEccByType(handle, counterType, NVML_MEMORY_ERROR_TYPE_UNCORRECTED))
     strResult += '        </double_bit>\n'
     return strResult
 
@@ -95,6 +116,43 @@ def GetEccStr(handle):
     strResult += '      </aggregate>\n'
     return strResult
 
+def StrGOM(mode):
+    if mode == NVML_GOM_ALL_ON:
+        return "All On";
+    elif mode == NVML_GOM_COMPUTE:
+        return "Compute";
+    elif mode == NVML_GOM_LOW_DP:
+        return "Low Double Precision";
+    else:
+        return "Unknown";
+
+def GetClocksThrottleReasons(handle):
+    throttleReasons = [
+            [nvmlClocksThrottleReasonGpuIdle,           "clocks_throttle_reason_gpu_idle"],
+            [nvmlClocksThrottleReasonUserDefinedClocks, "clocks_throttle_reason_user_defined_clocks"],
+            [nvmlClocksThrottleReasonSwPowerCap,        "clocks_throttle_reason_sw_power_cap"],
+            [nvmlClocksThrottleReasonHwSlowdown,        "clocks_throttle_reason_hw_slowdown"],
+            [nvmlClocksThrottleReasonUnknown,           "clocks_throttle_reason_unknown"]
+            ];
+
+    strResult = ''
+
+    try:
+        supportedClocksThrottleReasons = nvmlDeviceGetSupportedClocksThrottleReasons(handle);
+        clocksThrottleReasons = nvmlDeviceGetCurrentClocksThrottleReasons(handle);
+        strResult += '    <clocks_throttle_reasons>\n'
+        for (mask, name) in throttleReasons:
+            if (mask & supportedClocksThrottleReasons):
+                val = "Active" if mask & clocksThrottleReasons else "Not Active";
+            else:
+                val = handleError(NVML_ERROR_NOT_SUPPORTED);
+            strResult += "      <%s>%s</%s>\n" % (name, val, name);
+        strResult += '    </clocks_throttle_reasons>\n'
+    except NVMLError as err:
+        strResult += '    <clocks_throttle_reasons>%s</clocks_throttle_reasons>\n' % (handleError(err));
+
+    return strResult;
+        
 #
 # Converts errors into string messages
 #
@@ -115,7 +173,7 @@ def XmlDeviceQuery():
         strResult = ''
 
         strResult += '<?xml version="1.0" ?>\n'
-        strResult += '<!DOCTYPE nvidia_smi_log SYSTEM "nvsmi_device.dtd">\n'
+        strResult += '<!DOCTYPE nvidia_smi_log SYSTEM "nvsmi_device_v4.dtd">\n'
         strResult += '<nvidia_smi_log>\n'
 
         strResult += '  <timestamp>' + str(datetime.date.today()) + '</timestamp>\n'
@@ -139,7 +197,7 @@ def XmlDeviceQuery():
                 state = handleError(err)
             
             strResult += '    <display_mode>' + state + '</display_mode>\n'
-
+            
             try:
                 mode = 'Enabled' if (nvmlDeviceGetPersistenceMode(handle) != 0) else 'Disabled'
             except NVMLError as err:
@@ -150,13 +208,13 @@ def XmlDeviceQuery():
             strResult += '    <driver_model>\n'
 
             try:
-                current = str(nvmlDeviceGetCurrentDriverModel(handle))
+                current = 'WDDM' if (nvmlDeviceGetCurrentDriverModel(handle) == NVML_DRIVER_WDDM) else 'TCC' 
             except NVMLError as err:
                 current = handleError(err)
             strResult += '      <current_dm>' + current + '</current_dm>\n'
 
             try:
-                pending = str(nvmlDeviceGetPendingDriverModel(handle))
+                pending = 'WDDM' if (nvmlDeviceGetPendingDriverModel(handle) == NVML_DRIVER_WDDM) else 'TCC' 
             except NVMLError as err:
                 pending = handleError(err)
 
@@ -187,10 +245,15 @@ def XmlDeviceQuery():
 
             strResult += '    <inforom_version>\n'
             
+            try:
+                img = nvmlDeviceGetInforomImageVersion(handle)
+            except NVMLError as err:
+                img = handleError(err)
+                
+            strResult += '      <img_version>' + img + '</img_version>\n'
+
             try:
                 oem = nvmlDeviceGetInforomVersion(handle, NVML_INFOROM_OEM)
-                if oem == '':
-                    oem = 'N/A'
             except NVMLError as err:
                 oem = handleError(err)
                 
@@ -198,29 +261,44 @@ def XmlDeviceQuery():
             
             try:
                 ecc = nvmlDeviceGetInforomVersion(handle, NVML_INFOROM_ECC)
-                if ecc == '':
-                    ecc = 'N/A'
             except NVMLError as err:
                 ecc = handleError(err)
             
             strResult += '      <ecc_object>' + ecc + '</ecc_object>\n'
+
             try:
                 pwr = nvmlDeviceGetInforomVersion(handle, NVML_INFOROM_POWER)
-                if pwr == '':
-                    pwr = 'N/A'
             except NVMLError as err:
                 pwr = handleError(err)
             
             strResult += '      <pwr_object>' + pwr + '</pwr_object>\n'
+                       
             strResult += '    </inforom_version>\n'
 
+            strResult += '    <gpu_operation_mode>\n'
+
+            try:
+                current = StrGOM(nvmlDeviceGetCurrentGpuOperationMode(handle))
+            except NVMLError as err:
+                current = handleError(err)
+            strResult += '      <current_gom>' + current + '</current_gom>\n'
+
+            try:
+                pending = StrGOM(nvmlDeviceGetPendingGpuOperationMode(handle))
+            except NVMLError as err:
+                pending = handleError(err)
+
+            strResult += '      <pending_gom>' + pending + '</pending_gom>\n'
+
+            strResult += '    </gpu_operation_mode>\n'
+
             strResult += '    <pci>\n'
             strResult += '      <pci_bus>%02X</pci_bus>\n' % pciInfo.bus
             strResult += '      <pci_device>%02X</pci_device>\n' % pciInfo.device
             strResult += '      <pci_domain>%04X</pci_domain>\n' % pciInfo.domain
             strResult += '      <pci_device_id>%08X</pci_device_id>\n' % (pciInfo.pciDeviceId)
-            strResult += '      <pci_sub_system_id>%08X</pci_sub_system_id>\n' % (pciInfo.pciSubSystemId)
             strResult += '      <pci_bus_id>' + str(pciInfo.busId) + '</pci_bus_id>\n'
+            strResult += '      <pci_sub_system_id>%08X</pci_sub_system_id>\n' % (pciInfo.pciSubSystemId)
             strResult += '      <pci_gpu_link_info>\n'
 
 
@@ -266,11 +344,20 @@ def XmlDeviceQuery():
                 fan = handleError(err)
             strResult += '    <fan_speed>' + fan + '</fan_speed>\n'
 
+            try:
+                perfState = nvmlDeviceGetPowerState(handle)
+                perfStateStr = 'P%s' % perfState
+            except NVMLError as err:
+                perfStateStr = handleError(err)
+            strResult += '    <performance_state>' + perfStateStr + '</performance_state>\n'
+
+            strResult += GetClocksThrottleReasons(handle);
+
             try:
                 memInfo = nvmlDeviceGetMemoryInfo(handle)
                 mem_total = str(memInfo.total / 1024 / 1024) + ' MB'
                 mem_used = str(memInfo.used / 1024 / 1024) + ' MB'
-                mem_free = str(memInfo.free / 1024 / 1024) + ' MB'
+                mem_free = str(memInfo.total / 1024 / 1024 - memInfo.used / 1024 / 1024) + ' MB'
             except NVMLError as err:
                 error = handleError(err)
                 mem_total = error
@@ -303,16 +390,16 @@ def XmlDeviceQuery():
 
             try:
                 util = nvmlDeviceGetUtilizationRates(handle)
-                gpu_util = str(util.gpu)
-                mem_util = str(util.memory)
+                gpu_util = str(util.gpu) + ' %'
+                mem_util = str(util.memory) + ' %'
             except NVMLError as err:
                 error = handleError(err)
                 gpu_util = error
                 mem_util = error
             
             strResult += '    <utilization>\n'
-            strResult += '      <gpu_util>' + gpu_util + ' %</gpu_util>\n'
-            strResult += '      <memory_util>' + mem_util + ' %</memory_util>\n'
+            strResult += '      <gpu_util>' + gpu_util + '</gpu_util>\n'
+            strResult += '      <memory_util>' + mem_util + '</memory_util>\n'
             strResult += '    </utilization>\n'
             
             try:
@@ -344,10 +431,10 @@ def XmlDeviceQuery():
 
             strResult += '    <power_readings>\n'
             try:
-                perfState = nvmlDeviceGetPowerState(handle)
+                perfState = 'P' + str(nvmlDeviceGetPowerState(handle))
             except NVMLError as err:
                 perfState = handleError(err)
-            strResult += '      <power_state>P%s</power_state>\n' % perfState
+            strResult += '      <power_state>%s</power_state>\n' % perfState
             try:
                 powMan = nvmlDeviceGetPowerManagementMode(handle)
                 powManStr = 'Supported' if powMan != 0 else 'N/A'
@@ -362,10 +449,27 @@ def XmlDeviceQuery():
             strResult += '      <power_draw>' + powDrawStr + '</power_draw>\n'
             try:
                 powLimit = (nvmlDeviceGetPowerManagementLimit(handle) / 1000.0)
-                powLimitStr = '%d W' % powLimit
+                powLimitStr = '%.2f W' % powLimit
             except NVMLError as err:
                 powLimitStr = handleError(err)
             strResult += '      <power_limit>' + powLimitStr + '</power_limit>\n'
+            try:
+                powLimit = (nvmlDeviceGetPowerManagementDefaultLimit(handle) / 1000.0)
+                powLimitStr = '%.2f W' % powLimit
+            except NVMLError as err:
+                powLimitStr = handleError(err)
+            strResult += '      <default_power_limit>' + powLimitStr + '</default_power_limit>\n'
+            try:
+                powLimit = nvmlDeviceGetPowerManagementLimitConstraints(handle)
+                powLimitStrMin = '%.2f W' % (powLimit[0] / 1000.0)
+                powLimitStrMax = '%.2f W' % (powLimit[1] / 1000.0)
+            except NVMLError as err:
+                error = handleError(err)
+                powLimitStrMin = error
+                powLimitStrMax = error
+            strResult += '      <min_power_limit>' + powLimitStrMin + '</min_power_limit>\n'
+            strResult += '      <max_power_limit>' + powLimitStrMax + '</max_power_limit>\n'
+
             strResult += '    </power_readings>\n'
 
             strResult += '    <clocks>\n'
@@ -375,73 +479,96 @@ def XmlDeviceQuery():
                 graphics = handleError(err)
             strResult += '      <graphics_clock>' +graphics + ' MHz</graphics_clock>\n'
             try:
-                sm = str(nvmlDeviceGetClockInfo(handle, NVML_CLOCK_SM))
+                sm = str(nvmlDeviceGetClockInfo(handle, NVML_CLOCK_SM)) + ' MHz'
             except NVMLError as err:
                 sm = handleError(err)
-            strResult += '      <sm_clock>' + sm + ' MHz</sm_clock>\n'
+            strResult += '      <sm_clock>' + sm + '</sm_clock>\n'
             try:
-                mem = str(nvmlDeviceGetClockInfo(handle, NVML_CLOCK_MEM))
+                mem = str(nvmlDeviceGetClockInfo(handle, NVML_CLOCK_MEM)) + ' MHz'
             except NVMLError as err:
                 mem = handleError(err)
-            strResult += '      <mem_clock>' + mem + ' MHz</mem_clock>\n'
+            strResult += '      <mem_clock>' + mem + '</mem_clock>\n'
             strResult += '    </clocks>\n'
 
+            strResult += '    <applications_clocks>\n'
+            try:
+                graphics = str(nvmlDeviceGetApplicationsClock(handle, NVML_CLOCK_GRAPHICS)) + ' MHz'
+            except NVMLError as err:
+                graphics = handleError(err)
+            strResult += '      <graphics_clock>' +graphics + '</graphics_clock>\n'
+            try:
+                mem = str(nvmlDeviceGetApplicationsClock(handle, NVML_CLOCK_MEM)) + ' MHz'
+            except NVMLError as err:
+                mem = handleError(err)
+            strResult += '      <mem_clock>' + mem + '</mem_clock>\n'
+            strResult += '    </applications_clocks>\n'
+
             strResult += '    <max_clocks>\n'
             try:
-                graphics = str(nvmlDeviceGetMaxClockInfo(handle, NVML_CLOCK_GRAPHICS))
+                graphics = str(nvmlDeviceGetMaxClockInfo(handle, NVML_CLOCK_GRAPHICS)) + ' MHz'
             except NVMLError as err:
                 graphics = handleError(err)
-            strResult += '      <graphics_clock>' + graphics + ' MHz</graphics_clock>\n'
+            strResult += '      <graphics_clock>' + graphics + '</graphics_clock>\n'
             try:
-                sm = str(nvmlDeviceGetMaxClockInfo(handle, NVML_CLOCK_SM))
+                sm = str(nvmlDeviceGetMaxClockInfo(handle, NVML_CLOCK_SM)) + ' MHz'
             except NVMLError as err:
                 sm = handleError(err)
-            strResult += '      <sm_clock>' + sm + ' MHz</sm_clock>\n'
+            strResult += '      <sm_clock>' + sm + '</sm_clock>\n'
             try:
-                mem = str(nvmlDeviceGetMaxClockInfo(handle, NVML_CLOCK_MEM))
+                mem = str(nvmlDeviceGetMaxClockInfo(handle, NVML_CLOCK_MEM)) + ' MHz'
             except NVMLError as err:
                 mem = handleError(err)
-            strResult += '      <mem_clock>' + mem + ' MHz</mem_clock>\n'
+            strResult += '      <mem_clock>' + mem + '</mem_clock>\n'
             strResult += '    </max_clocks>\n'
-            
+
             try:
-                perfState = nvmlDeviceGetPowerState(handle)
-                perfStateStr = 'P%s' % perfState
+                memClocks = nvmlDeviceGetSupportedMemoryClocks(handle)
+                strResult += '    <supported_clocks>\n'
+
+                for m in memClocks:
+                    strResult += '      <supported_mem_clock>\n'
+                    strResult += '        <value>%d MHz</value>\n' % m
+                    try:
+                        clocks = nvmlDeviceGetSupportedGraphicsClocks(handle, m)
+                        for c in clocks:
+                            strResult += '        <supported_graphics_clock>%d MHz</supported_graphics_clock>\n' % c
+                    except NVMLError as err:
+                        strResult += '        <supported_graphics_clock>%s</supported_graphics_clock>\n' % handleError(err)
+                    strResult += '      </supported_mem_clock>\n'
+
+                strResult += '    </supported_clocks>\n'
             except NVMLError as err:
-                perfStateStr = handleError(err)
-            strResult += '    <performance_state>' + perfStateStr + '</performance_state>\n'
-            
-            strResult += '    <compute_processes>\n'
-            
-            procstr = ""
+                strResult += '    <supported_clocks>' + handleError(err) + '</supported_clocks>\n'
+
             try:
                 procs = nvmlDeviceGetComputeRunningProcesses(handle)
-            except NVMLError as err:
-                procs = []
-                procstr = handleError(err)
+                strResult += '    <compute_processes>\n'
              
-            for p in procs:
-                procstr += '    <process_info>\n'
-                procstr += '      <pid>%d</pid>\n' % p.pid
-                try:
-                    name = str(nvmlSystemGetProcessName(p.pid))
-                except NVMLError as err:
-                    if (err.value == NVML_ERROR_NOT_FOUND):
-                        # probably went away
-                        continue
+                for p in procs:
+                    try:
+                        name = str(nvmlSystemGetProcessName(p.pid))
+                    except NVMLError as err:
+                        if (err.value == NVML_ERROR_NOT_FOUND):
+                            # probably went away
+                            continue
+                        else:
+                            name = handleError(err)
+                    
+                    strResult += '    <process_info>\n'
+                    strResult += '      <pid>%d</pid>\n' % p.pid
+                    strResult += '      <process_name>' + name + '</process_name>\n'
+
+                    if (p.usedGpuMemory == None):
+                        mem = 'N\A'
                     else:
-                        name = handleError(err)
-                procstr += '      <process_name>' + name + '</process_name>\n'
-                procstr += '      <used_memory>\n'
-                if (p.usedGpuMemory == None):
-                    procstr += 'N\A'
-                else:
-                    procstr += '%d MB\n' % (p.usedGpuMemory / 1024 / 1024)
-                procstr += '</used_memory>\n'
-                procstr += '    </process_info>\n'
-            
-            strResult += procstr
-            strResult += '    </compute_processes>\n'
+                        mem = '%d MB' % (p.usedGpuMemory / 1024 / 1024)
+                    strResult += '      <used_memory>' + mem + '</used_memory>\n'
+                    strResult += '    </process_info>\n'
+                
+                strResult += '    </compute_processes>\n'
+            except NVMLError as err:
+                strResult += '    <compute_processes>' + handleError(err) + '</compute_processes>\n'
+
             strResult += '  </gpu>\n'
             
         strResult += '</nvidia_smi_log>\n'
